@@ -37,7 +37,7 @@ GST_DEBUG_CATEGORY_STATIC (ivorbisfile_debug);
   (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_IVORBISFILE,IvorbisfileClass))
 #define GST_IS_IVORBISFILE(obj)                                 \
   (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_IVORBISFILE))
-#define GST_IS_IVORBISFILE_CLASS(obj)                           \
+#define GST_IS_IVORBISFILE_CLASS(klass)                           \
   (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_IVORBISFILE))
 
 typedef struct _Ivorbisfile Ivorbisfile;
@@ -580,6 +580,11 @@ gst_ivorbisfile_sink_activate (GstPad * sinkpad)
     /* FIX ME */
     /* ivorbisfile->vf.seekable = TRUE; */
     ivorbisfile->vf.seekable = FALSE;
+    if (ivorbisfile->adapter) {
+      gst_adapter_clear (ivorbisfile->adapter);
+      g_object_unref (ivorbisfile->adapter);
+      ivorbisfile->adapter = NULL;
+    }
     return gst_pad_activate_pull (sinkpad, TRUE);
   } else {
 
