@@ -1,6 +1,7 @@
 /* GStreamer wavpack plugin
- * (c) 2005 Arwed v. Merkatz <v.merkatz@gmx.net>
- * (c) 2006 Tim-Philipp Müller <tim centricular net>
+ * Copyright (c) 2005 Arwed v. Merkatz <v.merkatz@gmx.net>
+ * Copyright (c) 2006 Tim-Philipp Müller <tim centricular net>
+ * Copyright (c) 2006 Sebastian Dröge <slomo@circular-chaos.org>
  *
  * gstwavpackparse.c: wavpack file parser
  *
@@ -20,6 +21,25 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/**
+ * SECTION:element-wavpackparse
+ *
+ * <refsect2>
+ * WavpackParse takes raw, unframed Wavpack streams and splits them into
+ * single Wavpack chunks with information like bit depth and the position
+ * in the stream.
+ * <ulink url="http://www.wavpack.com/">Wavpack</ulink> is an open-source
+ * audio codec that features both lossless and lossy encoding.
+ * <title>Example launch line</title>
+ * <para>
+ * <programlisting>
+ * gst-launch filesrc location=test.wv ! wavpackparse ! wavpackdec ! fakesink
+ * </programlisting>
+ * This pipeline decodes the Wavpack file test.wv into raw audio buffers.
+ * </para>
+ * </refsect2>
+ */
+
 #include <gst/gst.h>
 
 #include <math.h>
@@ -27,6 +47,7 @@
 
 #include <wavpack/wavpack.h>
 #include "gstwavpackparse.h"
+#include "gstwavpackstreamreader.h"
 #include "gstwavpackcommon.h"
 
 GST_DEBUG_CATEGORY_STATIC (gst_wavpack_parse_debug);
@@ -45,7 +66,7 @@ static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SOMETIMES,
     GST_STATIC_CAPS ("audio/x-wavpack, "
         "width = (int) { 8, 16, 24, 32 }, "
-        "channels = (int) { 1, 2 }, "
+        "channels = (int) [ 1, 2 ], "
         "rate = (int) [ 6000, 192000 ], " "framed = (boolean) true")
     );
 
@@ -55,17 +76,18 @@ static GstStaticPadTemplate wvc_src_factory = GST_STATIC_PAD_TEMPLATE ("wvcsrc",
     GST_STATIC_CAPS ("audio/x-wavpack-correction, " "framed = (boolean) true")
     );
 
-static gboolean gst_wavepack_parse_sink_activate (GstPad * sinkpad);
+static gboolean gst_wavpack_parse_sink_activate (GstPad * sinkpad);
 static gboolean
-gst_wavepack_parse_sink_activate_pull (GstPad * sinkpad, gboolean active);
+gst_wavpack_parse_sink_activate_pull (GstPad * sinkpad, gboolean active);
 
 static void gst_wavpack_parse_loop (GstElement * element);
 static GstStateChangeReturn gst_wavpack_parse_change_state (GstElement *
     element, GstStateChange transition);
-static void gst_wavpack_parse_reset (GstWavpackParse * wavpackparse);
+static void gst_wavpack_parse_reset (GstWavpackParse * parse);
 static gint64 gst_wavpack_parse_get_upstream_length (GstWavpackParse * wvparse);
 static GstBuffer *gst_wavpack_parse_pull_buffer (GstWavpackParse * wvparse,
     gint64 offset, guint size, GstFlowReturn * flow);
+static GstFlowReturn gst_wavpack_parse_chain (GstPad * pad, GstBuffer * buf);
 
 GST_BOILERPLATE (GstWavpackParse, gst_wavpack_parse, GstElement,
     GST_TYPE_ELEMENT);
@@ -77,7 +99,8 @@ gst_wavpack_parse_base_init (gpointer klass)
       GST_ELEMENT_DETAILS ("WavePack parser",
       "Codec/Demuxer/Audio",
       "Parses Wavpack files",
-      "Arwed v. Merkatz <v.merkatz@gmx.net>");
+      "Arwed v. Merkatz <v.merkatz@gmx.net>, "
+      "Sebastian Dröge <slomo@circular-chaos.org>");
   GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
 
   gst_element_class_add_pad_template (element_class,
@@ -90,10 +113,11 @@ gst_wavpack_parse_base_init (gpointer klass)
 }
 
 static void
-gst_wavpack_parse_dispose (GObject * object)
+gst_wavpack_parse_finalize (GObject * object)
 {
   gst_wavpack_parse_reset (GST_WAVPACK_PARSE (object));
-  G_OBJECT_CLASS (parent_class)->dispose (object);
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static void
@@ -105,7 +129,7 @@ gst_wavpack_parse_class_init (GstWavpackParseClass * klass)
   gobject_class = (GObjectClass *) klass;
   gstelement_class = (GstElementClass *) klass;
 
-  gobject_class->dispose = gst_wavpack_parse_dispose;
+  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_wavpack_parse_finalize);
   gstelement_class->change_state =
       GST_DEBUG_FUNCPTR (gst_wavpack_parse_change_state);
 }
@@ -137,13 +161,19 @@ gst_wavpack_parse_index_get_entry_from_sample (GstWavpackParse * wvparse,
     entry = &g_array_index (wvparse->entries, GstWavpackParseIndexEntry, i);
 
     GST_LOG_OBJECT (wvparse, "Index entry %03u: sample %" G_GINT64_FORMAT " @"
-        " byte %" G_GINT64_FORMAT, entry->sample_offset, entry->byte_offset);
+        " byte %" G_GINT64_FORMAT, i, entry->sample_offset, entry->byte_offset);
 
     if (entry->sample_offset <= sample_offset &&
         sample_offset < entry->sample_offset_end) {
       GST_LOG_OBJECT (wvparse, "found match");
       return entry;
     }
+
+    /* as the list is sorted and we first look at the latest entry
+     * we can abort searching for an entry if the sample we want is
+     * after the latest one */
+    if (sample_offset >= entry->sample_offset_end)
+      break;
   }
   GST_LOG_OBJECT (wvparse, "no match in index");
   return NULL;
@@ -177,39 +207,61 @@ gst_wavpack_parse_index_append_entry (GstWavpackParse * wvparse,
 }
 
 static void
-gst_wavpack_parse_reset (GstWavpackParse * wavpackparse)
+gst_wavpack_parse_reset (GstWavpackParse * parse)
 {
-  wavpackparse->total_samples = 0;
-  wavpackparse->samplerate = 0;
-  wavpackparse->channels = 0;
+  parse->total_samples = -1;
+  parse->samplerate = 0;
+  parse->channels = 0;
 
-  gst_segment_init (&wavpackparse->segment, GST_FORMAT_UNDEFINED);
+  gst_segment_init (&parse->segment, GST_FORMAT_UNDEFINED);
 
-  wavpackparse->current_offset = 0;
-  wavpackparse->need_newsegment = TRUE;
-  wavpackparse->upstream_length = -1;
+  parse->current_offset = 0;
+  parse->need_newsegment = TRUE;
+  parse->upstream_length = -1;
 
-  if (wavpackparse->entries) {
-    g_array_free (wavpackparse->entries, TRUE);
-    wavpackparse->entries = NULL;
+  if (parse->entries) {
+    g_array_free (parse->entries, TRUE);
+    parse->entries = NULL;
   }
 
-  if (wavpackparse->srcpad != NULL) {
+  if (parse->adapter) {
+    gst_adapter_clear (parse->adapter);
+    g_object_unref (parse->adapter);
+    parse->adapter = NULL;
+  }
+
+  if (parse->srcpad != NULL) {
     gboolean res;
 
-    GST_DEBUG_OBJECT (wavpackparse, "Removing src pad");
-    res = gst_element_remove_pad (GST_ELEMENT (wavpackparse),
-        wavpackparse->srcpad);
+    GST_DEBUG_OBJECT (parse, "Removing src pad");
+    res = gst_element_remove_pad (GST_ELEMENT (parse), parse->srcpad);
     g_return_if_fail (res != FALSE);
-    gst_object_unref (wavpackparse->srcpad);
-    wavpackparse->srcpad = NULL;
+    gst_object_unref (parse->srcpad);
+    parse->srcpad = NULL;
   }
+
+  g_list_foreach (parse->queued_events, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (parse->queued_events);
+  parse->queued_events = NULL;
+}
+
+static const GstQueryType *
+gst_wavpack_parse_get_src_query_types (GstPad * pad)
+{
+  static const GstQueryType types[] = {
+    GST_QUERY_POSITION,
+    GST_QUERY_DURATION,
+    GST_QUERY_SEEKING,
+    0
+  };
+
+  return types;
 }
 
 static gboolean
 gst_wavpack_parse_src_query (GstPad * pad, GstQuery * query)
 {
-  GstWavpackParse *wavpackparse = GST_WAVPACK_PARSE (gst_pad_get_parent (pad));
+  GstWavpackParse *parse = GST_WAVPACK_PARSE (gst_pad_get_parent (pad));
   GstFormat format;
   gboolean ret = FALSE;
 
@@ -218,14 +270,14 @@ gst_wavpack_parse_src_query (GstPad * pad, GstQuery * query)
       gint64 cur, len;
       guint rate;
 
-      GST_OBJECT_LOCK (wavpackparse);
-      cur = wavpackparse->segment.last_stop;
-      len = wavpackparse->total_samples;
-      rate = wavpackparse->samplerate;
-      GST_OBJECT_UNLOCK (wavpackparse);
+      GST_OBJECT_LOCK (parse);
+      cur = parse->segment.last_stop;
+      len = parse->total_samples;
+      rate = parse->samplerate;
+      GST_OBJECT_UNLOCK (parse);
 
       if (len <= 0 || rate == 0) {
-        GST_DEBUG_OBJECT (wavpackparse, "haven't read header yet");
+        GST_DEBUG_OBJECT (parse, "haven't read header yet");
         break;
       }
 
@@ -242,8 +294,9 @@ gst_wavpack_parse_src_query (GstPad * pad, GstQuery * query)
           ret = TRUE;
           break;
         default:
-          GST_DEBUG_OBJECT (wavpackparse, "cannot handle position query in "
-              "%s format", gst_format_get_name (format));
+          GST_DEBUG_OBJECT (parse, "cannot handle position query in "
+              "%s format. Forwarding upstream.", gst_format_get_name (format));
+          ret = gst_pad_query_default (pad, query);
           break;
       }
       break;
@@ -252,13 +305,15 @@ gst_wavpack_parse_src_query (GstPad * pad, GstQuery * query)
       gint64 len;
       guint rate;
 
-      GST_OBJECT_LOCK (wavpackparse);
-      rate = wavpackparse->samplerate;
-      len = wavpackparse->total_samples;
-      GST_OBJECT_UNLOCK (wavpackparse);
+      GST_OBJECT_LOCK (parse);
+      rate = parse->samplerate;
+      /* FIXME: return 0 if we work in push based mode to let totem
+       * recognize that we can't seek */
+      len = (parse->adapter) ? 0 : parse->total_samples;
+      GST_OBJECT_UNLOCK (parse);
 
-      if (len <= 0 || rate == 0) {
-        GST_DEBUG_OBJECT (wavpackparse, "haven't read header yet");
+      if (len < 0 || rate == 0) {
+        GST_DEBUG_OBJECT (parse, "haven't read header yet");
         break;
       }
 
@@ -275,19 +330,38 @@ gst_wavpack_parse_src_query (GstPad * pad, GstQuery * query)
           ret = TRUE;
           break;
         default:
-          GST_DEBUG_OBJECT (wavpackparse, "cannot handle duration query in "
-              "%s format", gst_format_get_name (format));
+          GST_DEBUG_OBJECT (parse, "cannot handle duration query in "
+              "%s format. Forwarding upstream.", gst_format_get_name (format));
+          ret = gst_pad_query_default (pad, query);
           break;
       }
       break;
     }
+    case GST_QUERY_SEEKING:{
+      gst_query_parse_seeking (query, &format, NULL, NULL, NULL);
+      if (format == GST_FORMAT_TIME || format == GST_FORMAT_DEFAULT) {
+        gboolean seekable;
+        gint64 duration = -1;
+
+        gst_pad_query_duration (pad, &format, &duration);
+
+        /* can't seek in streaming mode yet */
+        GST_OBJECT_LOCK (parse);
+        seekable = (parse->adapter != NULL);
+        GST_OBJECT_UNLOCK (parse);
+
+        gst_query_set_seeking (query, GST_FORMAT_TIME, seekable, 0, duration);
+        ret = TRUE;
+      }
+      break;
+    }
     default:{
       ret = gst_pad_query_default (pad, query);
       break;
     }
   }
 
-  gst_object_unref (wavpackparse);
+  gst_object_unref (parse);
   return ret;
 
 }
@@ -331,9 +405,7 @@ gst_wavpack_parse_scan_to_find_sample (GstWavpackParse * parse,
 
   /* now scan forward until we find the chunk we're looking for or hit EOS */
   do {
-    WavpackHeader header = { {0,}
-    , 0,
-    };
+    WavpackHeader header;
     GstBuffer *buf;
 
     buf = gst_wavpack_parse_pull_buffer (parse, off, sizeof (WavpackHeader),
@@ -421,6 +493,11 @@ gst_wavpack_parse_handle_seek_event (GstWavpackParse * wvparse,
   gint64 chunk_start;           /* first sample in chunk we seek to           */
   guint rate;
 
+  if (wvparse->adapter) {
+    GST_DEBUG_OBJECT (wvparse, "seeking in streaming mode not implemented yet");
+    return FALSE;
+  }
+
   gst_event_parse_seek (event, &speed, &format, &seek_flags, &start_type,
       &start, &stop_type, &stop);
 
@@ -451,6 +528,11 @@ gst_wavpack_parse_handle_seek_event (GstWavpackParse * wvparse,
       stop = gst_util_uint64_scale_int (stop, rate, GST_SECOND);
   }
 
+  /* if seek is to something after the end of the stream seek only
+   * to the end. this can be caused by rounding errors */
+  if (start >= wvparse->total_samples)
+    start = wvparse->total_samples;
+
   flush = ((seek_flags & GST_SEEK_FLAG_FLUSH) != 0);
 
   if (start < 0) {
@@ -519,69 +601,111 @@ gst_wavpack_parse_handle_seek_event (GstWavpackParse * wvparse,
   return ret;
 }
 
+static gboolean
+gst_wavpack_parse_sink_event (GstPad * pad, GstEvent * event)
+{
+  GstWavpackParse *parse;
+  gboolean ret = TRUE;
+
+  parse = GST_WAVPACK_PARSE (gst_pad_get_parent (pad));
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_FLUSH_STOP:{
+      if (parse->adapter) {
+        gst_adapter_clear (parse->adapter);
+      }
+      ret = gst_pad_push_event (parse->srcpad, event);
+      break;
+    }
+    case GST_EVENT_NEWSEGMENT:{
+      parse->need_newsegment = TRUE;
+      gst_event_unref (event);
+      ret = TRUE;
+      break;
+    }
+    case GST_EVENT_EOS:{
+      if (parse->adapter) {
+        /* remove all bytes that are left in the adapter after EOS. They can't
+         * be a complete Wavpack block and we can't do anything with them */
+        gst_adapter_clear (parse->adapter);
+      }
+      ret = gst_pad_push_event (parse->srcpad, event);
+      break;
+    }
+    default:{
+      /* stream lock is recursive, should be fine for all events */
+      GST_PAD_STREAM_LOCK (pad);
+      if (parse->srcpad == NULL) {
+        parse->queued_events = g_list_append (parse->queued_events, event);
+      } else {
+        ret = gst_pad_push_event (parse->srcpad, event);
+      }
+      GST_PAD_STREAM_UNLOCK (pad);
+    }
+  }
+
+
+  gst_object_unref (parse);
+  return ret;
+}
+
 static gboolean
 gst_wavpack_parse_src_event (GstPad * pad, GstEvent * event)
 {
-  GstWavpackParse *wavpackparse;
+  GstWavpackParse *parse;
   gboolean ret;
 
-  wavpackparse = GST_WAVPACK_PARSE (gst_pad_get_parent (pad));
+  parse = GST_WAVPACK_PARSE (gst_pad_get_parent (pad));
 
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_SEEK:
-      ret = gst_wavpack_parse_handle_seek_event (wavpackparse, event);
+      ret = gst_wavpack_parse_handle_seek_event (parse, event);
       break;
     default:
       ret = gst_pad_event_default (pad, event);
       break;
   }
 
-  gst_object_unref (wavpackparse);
+  gst_object_unref (parse);
   return ret;
 }
 
 static void
-gst_wavpack_parse_init (GstWavpackParse * wavpackparse,
-    GstWavpackParseClass * gclass)
+gst_wavpack_parse_init (GstWavpackParse * parse, GstWavpackParseClass * gclass)
 {
-  GstElementClass *klass = GST_ELEMENT_GET_CLASS (wavpackparse);
+  GstElementClass *klass = GST_ELEMENT_GET_CLASS (parse);
   GstPadTemplate *tmpl;
 
   tmpl = gst_element_class_get_pad_template (klass, "sink");
-  wavpackparse->sinkpad = gst_pad_new_from_template (tmpl, "sink");
+  parse->sinkpad = gst_pad_new_from_template (tmpl, "sink");
 
-  gst_pad_set_activate_function (wavpackparse->sinkpad,
-      GST_DEBUG_FUNCPTR (gst_wavepack_parse_sink_activate));
+  gst_pad_set_activate_function (parse->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_wavpack_parse_sink_activate));
+  gst_pad_set_activatepull_function (parse->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_wavpack_parse_sink_activate_pull));
+  gst_pad_set_event_function (parse->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_wavpack_parse_sink_event));
+  gst_pad_set_chain_function (parse->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_wavpack_parse_chain));
 
-  gst_pad_set_activatepull_function (wavpackparse->sinkpad,
-      GST_DEBUG_FUNCPTR (gst_wavepack_parse_sink_activate_pull));
+  gst_element_add_pad (GST_ELEMENT (parse), parse->sinkpad);
 
-  gst_element_add_pad (GST_ELEMENT (wavpackparse), wavpackparse->sinkpad);
-
-  wavpackparse->srcpad = NULL;
-  gst_wavpack_parse_reset (wavpackparse);
+  parse->srcpad = NULL;
+  gst_wavpack_parse_reset (parse);
 }
 
 static gint64
-gst_wavpack_parse_get_upstream_length (GstWavpackParse * wavpackparse)
+gst_wavpack_parse_get_upstream_length (GstWavpackParse * parse)
 {
-  GstPad *peer;
   gint64 length = -1;
 
-  peer = gst_pad_get_peer (wavpackparse->sinkpad);
-  if (peer) {
-    GstFormat format = GST_FORMAT_BYTES;
+  GstFormat format = GST_FORMAT_BYTES;
 
-    if (!gst_pad_query_duration (peer, &format, &length)) {
-      length = -1;
-    } else {
-      GST_DEBUG ("upstream length: %" G_GINT64_FORMAT, length);
-    }
-    gst_object_unref (peer);
+  if (!gst_pad_query_peer_duration (parse->sinkpad, &format, &length)) {
+    length = -1;
   } else {
-    GST_DEBUG ("no peer!");
+    GST_DEBUG ("upstream length: %" G_GINT64_FORMAT, length);
   }
-
   return length;
 }
 
@@ -628,7 +752,7 @@ static gboolean
 gst_wavpack_parse_create_src_pad (GstWavpackParse * wvparse, GstBuffer * buf,
     WavpackHeader * header)
 {
-  WavpackMetadata meta;
+  GstWavpackMetadata meta;
   GstCaps *caps = NULL;
   guchar *bufptr;
 
@@ -636,7 +760,7 @@ gst_wavpack_parse_create_src_pad (GstWavpackParse * wvparse, GstBuffer * buf,
 
   bufptr = GST_BUFFER_DATA (buf) + sizeof (WavpackHeader);
 
-  while (read_metadata_buff (&meta, GST_BUFFER_DATA (buf), &bufptr)) {
+  while (gst_wavpack_read_metadata (&meta, GST_BUFFER_DATA (buf), &bufptr)) {
     switch (meta.id) {
       case ID_WVC_BITSTREAM:{
         caps = gst_caps_new_simple ("audio/x-wavpack-correction",
@@ -646,23 +770,41 @@ gst_wavpack_parse_create_src_pad (GstWavpackParse * wvparse, GstBuffer * buf,
             (GST_ELEMENT_GET_CLASS (wvparse), "wvcsrc"), "wvcsrc");
         break;
       }
-      case ID_RIFF_HEADER:{
-        WaveHeader wheader;
+      case ID_WV_BITSTREAM:
+      case ID_WVX_BITSTREAM:{
+        WavpackStreamReader *stream_reader = gst_wavpack_stream_reader_new ();
+        WavpackContext *wpc;
+        gchar error_msg[80];
+        read_id rid;
 
-        /* skip RiffChunkHeader and ChunkHeader */
-        g_memmove (&wheader, meta.data + 20, sizeof (WaveHeader));
-        little_endian_to_native (&wheader, WaveHeaderFormat);
-        wvparse->samplerate = wheader.SampleRate;
-        wvparse->channels = wheader.NumChannels;
+        rid.buffer = GST_BUFFER_DATA (buf);
+        rid.length = GST_BUFFER_SIZE (buf);
+        rid.position = 0;
+
+        wpc =
+            WavpackOpenFileInputEx (stream_reader, &rid, NULL, error_msg, 0, 0);
+
+        if (!wpc)
+          return FALSE;
+
+        wvparse->samplerate = WavpackGetSampleRate (wpc);
+        wvparse->channels = WavpackGetNumChannels (wpc);
         wvparse->total_samples = header->total_samples;
+        if (wvparse->total_samples == (int32_t) - 1)
+          wvparse->total_samples = 0;
+        else
+          wvparse->total_samples--;
+
         caps = gst_caps_new_simple ("audio/x-wavpack",
-            "width", G_TYPE_INT, wheader.BitsPerSample,
+            "width", G_TYPE_INT, WavpackGetBitsPerSample (wpc),
             "channels", G_TYPE_INT, wvparse->channels,
             "rate", G_TYPE_INT, wvparse->samplerate,
             "framed", G_TYPE_BOOLEAN, TRUE, NULL);
         wvparse->srcpad =
             gst_pad_new_from_template (gst_element_class_get_pad_template
             (GST_ELEMENT_GET_CLASS (wvparse), "src"), "src");
+        WavpackCloseFile (wpc);
+        g_free (stream_reader);
         break;
       }
       default:{
@@ -681,32 +823,164 @@ gst_wavpack_parse_create_src_pad (GstWavpackParse * wvparse, GstBuffer * buf,
 
   gst_pad_set_query_function (wvparse->srcpad,
       GST_DEBUG_FUNCPTR (gst_wavpack_parse_src_query));
+  gst_pad_set_query_type_function (wvparse->srcpad,
+      GST_DEBUG_FUNCPTR (gst_wavpack_parse_get_src_query_types));
   gst_pad_set_event_function (wvparse->srcpad,
       GST_DEBUG_FUNCPTR (gst_wavpack_parse_src_event));
 
   gst_pad_set_caps (wvparse->srcpad, caps);
+  gst_caps_unref (caps);
   gst_pad_use_fixed_caps (wvparse->srcpad);
 
   gst_object_ref (wvparse->srcpad);
+  gst_pad_set_active (wvparse->srcpad, TRUE);
   gst_element_add_pad (GST_ELEMENT (wvparse), wvparse->srcpad);
   gst_element_no_more_pads (GST_ELEMENT (wvparse));
 
   return TRUE;
 }
 
+static GstFlowReturn
+gst_wavpack_parse_push_buffer (GstWavpackParse * wvparse, GstBuffer * buf,
+    WavpackHeader * header)
+{
+  wvparse->current_offset += header->ckSize + 8;
+
+  wvparse->segment.last_stop = header->block_index;
+
+  if (wvparse->need_newsegment) {
+    if (gst_wavpack_parse_send_newsegment (wvparse, FALSE))
+      wvparse->need_newsegment = FALSE;
+  }
+
+  /* send any queued events */
+  if (wvparse->queued_events) {
+    GList *l;
+
+    for (l = wvparse->queued_events; l != NULL; l = l->next) {
+      gst_pad_push_event (wvparse->srcpad, GST_EVENT (l->data));
+    }
+    g_list_free (wvparse->queued_events);
+    wvparse->queued_events = NULL;
+  }
+
+  GST_BUFFER_TIMESTAMP (buf) = gst_util_uint64_scale_int (header->block_index,
+      GST_SECOND, wvparse->samplerate);
+  GST_BUFFER_DURATION (buf) = gst_util_uint64_scale_int (header->block_samples,
+      GST_SECOND, wvparse->samplerate);
+  GST_BUFFER_OFFSET (buf) = header->block_index;
+  gst_buffer_set_caps (buf, GST_PAD_CAPS (wvparse->srcpad));
+
+  GST_LOG_OBJECT (wvparse, "Pushing buffer with time %" GST_TIME_FORMAT,
+      GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)));
+
+  return gst_pad_push (wvparse->srcpad, buf);
+}
+
+static guint8 *
+gst_wavpack_parse_find_marker (guint8 * buf, guint size)
+{
+  int i;
+  guint8 *ret = NULL;
+
+  if (G_UNLIKELY (size < 4))
+    return NULL;
+
+  for (i = 0; i < size - 4; i++) {
+    if (memcmp (buf + i, "wvpk", 4) == 0) {
+      ret = buf + i;
+      break;
+    }
+  }
+  return ret;
+}
+
+static GstFlowReturn
+gst_wavpack_parse_resync_loop (GstWavpackParse * parse, WavpackHeader * header)
+{
+  GstFlowReturn flow_ret = GST_FLOW_UNEXPECTED;
+  GstBuffer *buf = NULL;
+
+  /* loop until we have a frame header or reach the end of the stream */
+  while (1) {
+    guint8 *data, *marker;
+    guint len, size;
+
+    if (buf) {
+      gst_buffer_unref (buf);
+      buf = NULL;
+    }
+
+    if (parse->upstream_length == 0 ||
+        parse->upstream_length <= parse->current_offset) {
+      parse->upstream_length = gst_wavpack_parse_get_upstream_length (parse);
+      if (parse->upstream_length == 0 ||
+          parse->upstream_length <= parse->current_offset) {
+        break;
+      }
+    }
+
+    len = MIN (parse->upstream_length - parse->current_offset, 2048);
+
+    GST_LOG_OBJECT (parse, "offset: %" G_GINT64_FORMAT, parse->current_offset);
+
+    buf = gst_wavpack_parse_pull_buffer (parse, parse->current_offset,
+        len, &flow_ret);
+
+    /* whatever the problem is, there's nothing more for us to do for now */
+    if (buf == NULL)
+      break;
+
+    data = GST_BUFFER_DATA (buf);
+    size = GST_BUFFER_SIZE (buf);
+
+    /* not enough data for a header? */
+    if (size < sizeof (WavpackHeader))
+      break;
+
+    /* got a header right where we are at now? */
+    if (gst_wavpack_read_header (header, data))
+      break;
+
+    /* nope, let's see if we can find one */
+    marker = gst_wavpack_parse_find_marker (data + 1, size - 1);
+
+    if (marker) {
+      parse->current_offset += marker - data;
+      /* do one more loop iteration to make sure we pull enough
+       * data for a full header, we'll bail out then */
+    } else {
+      parse->current_offset += len - 4;
+    }
+  }
+
+  if (buf)
+    gst_buffer_unref (buf);
+
+  return flow_ret;
+}
+
 static void
 gst_wavpack_parse_loop (GstElement * element)
 {
-  GstWavpackParse *wavpackparse = GST_WAVPACK_PARSE (element);
+  GstWavpackParse *parse = GST_WAVPACK_PARSE (element);
   GstFlowReturn flow_ret;
   WavpackHeader header = { {0,}, 0, };
   GstBuffer *buf = NULL;
 
-  GST_LOG_OBJECT (wavpackparse, "Current offset: %" G_GINT64_FORMAT,
-      wavpackparse->current_offset);
+  flow_ret = gst_wavpack_parse_resync_loop (parse, &header);
 
-  buf = gst_wavpack_parse_pull_buffer (wavpackparse,
-      wavpackparse->current_offset, sizeof (WavpackHeader), &flow_ret);
+  if (flow_ret == GST_FLOW_UNEXPECTED) {
+    goto eos;
+  } else if (flow_ret != GST_FLOW_OK) {
+    goto pause;
+  }
+
+  GST_LOG_OBJECT (parse, "Read header at offset %" G_GINT64_FORMAT
+      ": chunk size = %u+8", parse->current_offset, header.ckSize);
+
+  buf = gst_wavpack_parse_pull_buffer (parse, parse->current_offset,
+      header.ckSize + 8, &flow_ret);
 
   if (buf == NULL && flow_ret == GST_FLOW_UNEXPECTED) {
     goto eos;
@@ -714,53 +988,19 @@ gst_wavpack_parse_loop (GstElement * element)
     goto pause;
   }
 
-  gst_wavpack_read_header (&header, GST_BUFFER_DATA (buf));
-  gst_buffer_unref (buf);
-
-  GST_LOG_OBJECT (wavpackparse, "Read header at offset %" G_GINT64_FORMAT
-      ": chunk size = %u+8", wavpackparse->current_offset, header.ckSize);
-
-  buf = gst_wavpack_parse_pull_buffer (wavpackparse,
-      wavpackparse->current_offset, header.ckSize + 8, &flow_ret);
-
-  if (buf == NULL && flow_ret == GST_FLOW_UNEXPECTED) {
-    goto eos;
-  } else if (buf == NULL) {
-    goto pause;
-  }
-
-  if (wavpackparse->srcpad == NULL) {
-    if (!gst_wavpack_parse_create_src_pad (wavpackparse, buf, &header)) {
-      GST_ELEMENT_ERROR (wavpackparse, STREAM, DECODE, (NULL), (NULL));
+  if (parse->srcpad == NULL) {
+    if (!gst_wavpack_parse_create_src_pad (parse, buf, &header)) {
+      GST_ELEMENT_ERROR (parse, STREAM, DECODE, (NULL), (NULL));
       goto pause;
     }
   }
 
-  gst_wavpack_parse_index_append_entry (wavpackparse,
-      wavpackparse->current_offset, header.block_index, header.block_samples);
+  gst_wavpack_parse_index_append_entry (parse, parse->current_offset,
+      header.block_index, header.block_samples);
 
-  wavpackparse->current_offset += header.ckSize + 8;
-
-  wavpackparse->segment.last_stop = header.block_index;
-
-  if (wavpackparse->need_newsegment) {
-    if (gst_wavpack_parse_send_newsegment (wavpackparse, FALSE))
-      wavpackparse->need_newsegment = FALSE;
-  }
-
-  GST_BUFFER_TIMESTAMP (buf) = gst_util_uint64_scale_int (header.block_index,
-      GST_SECOND, wavpackparse->samplerate);
-  GST_BUFFER_DURATION (buf) = gst_util_uint64_scale_int (header.block_samples,
-      GST_SECOND, wavpackparse->samplerate);
-  GST_BUFFER_OFFSET (buf) = header.block_index;
-  gst_buffer_set_caps (buf, GST_PAD_CAPS (wavpackparse->srcpad));
-
-  GST_LOG_OBJECT (wavpackparse, "Pushing buffer with time %" GST_TIME_FORMAT,
-      GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)));
-
-  flow_ret = gst_pad_push (wavpackparse->srcpad, buf);
+  flow_ret = gst_wavpack_parse_push_buffer (parse, buf, &header);
   if (flow_ret != GST_FLOW_OK) {
-    GST_DEBUG_OBJECT (wavpackparse, "Push failed, flow: %s",
+    GST_DEBUG_OBJECT (parse, "Push failed, flow: %s",
         gst_flow_get_name (flow_ret));
     goto pause;
   }
@@ -769,20 +1009,111 @@ gst_wavpack_parse_loop (GstElement * element)
 
 eos:
   {
-    GST_DEBUG_OBJECT (wavpackparse, "sending EOS");
-    if (wavpackparse->srcpad) {
-      gst_pad_push_event (wavpackparse->srcpad, gst_event_new_eos ());
+    GST_DEBUG_OBJECT (parse, "sending EOS");
+    if (parse->srcpad) {
+      gst_pad_push_event (parse->srcpad, gst_event_new_eos ());
     }
     /* fall through and pause task */
   }
 pause:
   {
-    GST_DEBUG_OBJECT (wavpackparse, "Pausing task");
-    gst_pad_pause_task (wavpackparse->sinkpad);
+    GST_DEBUG_OBJECT (parse, "Pausing task");
+    gst_pad_pause_task (parse->sinkpad);
     return;
   }
 }
 
+static gboolean
+gst_wavpack_parse_resync_adapter (GstAdapter * adapter)
+{
+  const guint8 *buf, *marker;
+  guint avail = gst_adapter_available (adapter);
+
+  if (avail < 4)
+    return FALSE;
+
+  /* if the marker is at the beginning don't do the expensive search */
+  buf = gst_adapter_peek (adapter, 4);
+  if (memcmp (buf, "wvpk", 4) == 0)
+    return TRUE;
+
+  if (avail == 4)
+    return FALSE;
+
+  /* search for the marker in the complete content of the adapter */
+  buf = gst_adapter_peek (adapter, avail);
+  if (buf && (marker = gst_wavpack_parse_find_marker ((guint8 *) buf, avail))) {
+    gst_adapter_flush (adapter, marker - buf);
+    return TRUE;
+  }
+
+  /* flush everything except the last 4 bytes. they could contain
+   * the start of a new marker */
+  gst_adapter_flush (adapter, avail - 4);
+
+  return FALSE;
+}
+
+static GstFlowReturn
+gst_wavpack_parse_chain (GstPad * pad, GstBuffer * buf)
+{
+  GstWavpackParse *wvparse = GST_WAVPACK_PARSE (GST_PAD_PARENT (pad));
+  GstFlowReturn ret = GST_FLOW_OK;
+  WavpackHeader wph;
+  const guint8 *tmp_buf;
+
+  if (!wvparse->adapter) {
+    wvparse->adapter = gst_adapter_new ();
+  }
+
+  if (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DISCONT)) {
+    gst_adapter_clear (wvparse->adapter);
+  }
+
+  gst_adapter_push (wvparse->adapter, buf);
+
+  if (gst_adapter_available (wvparse->adapter) < sizeof (WavpackHeader))
+    return ret;
+
+  if (!gst_wavpack_parse_resync_adapter (wvparse->adapter))
+    return ret;
+
+  tmp_buf = gst_adapter_peek (wvparse->adapter, sizeof (WavpackHeader));
+  gst_wavpack_read_header (&wph, (guint8 *) tmp_buf);
+
+  while (gst_adapter_available (wvparse->adapter) >= wph.ckSize + 4 * 1 + 4) {
+    GstBuffer *outbuf =
+        gst_adapter_take_buffer (wvparse->adapter, wph.ckSize + 4 * 1 + 4);
+
+    if (!outbuf)
+      return GST_FLOW_ERROR;
+
+    if (wvparse->srcpad == NULL) {
+      if (!gst_wavpack_parse_create_src_pad (wvparse, outbuf, &wph)) {
+        GST_ELEMENT_ERROR (wvparse, STREAM, DECODE, (NULL), (NULL));
+        ret = GST_FLOW_ERROR;
+        break;
+      }
+    }
+
+    ret = gst_wavpack_parse_push_buffer (wvparse, outbuf, &wph);
+
+    if (ret != GST_FLOW_OK)
+      break;
+
+    if (gst_adapter_available (wvparse->adapter) >= sizeof (WavpackHeader)) {
+      tmp_buf = gst_adapter_peek (wvparse->adapter, sizeof (WavpackHeader));
+
+      if (!gst_wavpack_parse_resync_adapter (wvparse->adapter))
+        break;
+
+      gst_wavpack_read_header (&wph, (guint8 *) tmp_buf);
+    }
+  }
+
+  return ret;
+}
+
 static GstStateChangeReturn
 gst_wavpack_parse_change_state (GstElement * element, GstStateChange transition)
 {
@@ -811,19 +1142,18 @@ gst_wavpack_parse_change_state (GstElement * element, GstStateChange transition)
   return ret;
 }
 
-
 static gboolean
-gst_wavepack_parse_sink_activate (GstPad * sinkpad)
+gst_wavpack_parse_sink_activate (GstPad * sinkpad)
 {
   if (gst_pad_check_pull_range (sinkpad)) {
     return gst_pad_activate_pull (sinkpad, TRUE);
   } else {
-    return FALSE;
+    return gst_pad_activate_push (sinkpad, TRUE);
   }
 }
 
 static gboolean
-gst_wavepack_parse_sink_activate_pull (GstPad * sinkpad, gboolean active)
+gst_wavpack_parse_sink_activate_pull (GstPad * sinkpad, gboolean active)
 {
   gboolean result;
 
